<?php declare(strict_types=1);

namespace June\Service\ScheduledTask;

use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use Psr\Log\LoggerInterface;
use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\OrFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Sorting\FieldSorting;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTask;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskDefinition;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskEntity;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskHandler;
use Shopware\Core\System\SalesChannel\SalesChannelEntity;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use June\Subscriber\JuneCommerceSubscriber;

use Shopware\Core\System\Locale\LocaleEntity;
use Shopware\Core\System\Country\CountryEntity;
use Shopware\Core\System\Tag\TagEntity;
use Symfony\Component\Messenger\Attribute\AsMessageHandler;

#[AsMessageHandler(handles: ExportCustomerUpdatesTask::class)]
class ExportCustomerUpdatesTaskHandler extends ScheduledTaskHandler {

    const CUSTOMER_LIMIT_PER_REQUEST = 200;
    const CONF_CURSOR_DATE_POSITION = 'June.config.last_customer_updated_at_cursor';
    const CONF_CURSOR_PAGINATION_POSITION = 'June.config.last_customer_pagination_cursor';

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $customerRepository;

    private EntityRepository $newsletterRecipientRepository;

    private EntityRepository $customerTagsRepository;

    private EntityRepository $salesChannelRepository;

    private EntityRepository $juneCustomerRepository;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private ?string $lastExecutionTime = null;

    private int $maxExecutionTime = 45;

    private array $loadedSubscriberTags = [];

    public function __construct
    (
        EntityRepository $scheduledTaskRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $customerRepository,
        EntityRepository $newsletterRecipientRepository,
        EntityRepository $customerTagsRepository,
        EntityRepository $salesChannelRepository,
        EntityRepository $juneCustomerRepository,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient,
        LoggerInterface $logger,
    )
    {
        parent::__construct($scheduledTaskRepository, $logger);
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->customerRepository = $customerRepository;
        $this->newsletterRecipientRepository = $newsletterRecipientRepository;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->salesChannelRepository = $salesChannelRepository;
        $this->juneCustomerRepository = $juneCustomerRepository;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;

        $serverMaxExecutionTime = ini_get('max_execution_time');
        if ($serverMaxExecutionTime && $serverMaxExecutionTime > 0) {
            $this->maxExecutionTime = $serverMaxExecutionTime - 5;
        }
        if ($this->maxExecutionTime <= 0) {
            $this->maxExecutionTime = $serverMaxExecutionTime * 1;
        }

        $this->lastExecutionTime = null;
    }

    /**
     * @param ScheduledTask $task
     * @return void
     */
    public function __invoke(ScheduledTask $task): void
    {
        $taskId = $task->getTaskId();
        if ($taskId === null) {
            // run task independent of the schedule
            $this->run();
            return;
        }

        /** @var ScheduledTaskEntity|null $taskEntity */
        $taskEntity = $this->scheduledTaskRepository
            ->search(new Criteria([$taskId]), Context::createDefaultContext())
            ->get($taskId);

        if ($taskEntity === null || !$taskEntity->isExecutionAllowed()) {
            return;
        }

        $this->markTaskRunning($task);
        try {
            $lastTaskExecution = $taskEntity->getLastExecutionTime();

            if ($this->configService->get(self::CONF_CURSOR_DATE_POSITION)) {
                $this->lastExecutionTime = $this->configService->get(self::CONF_CURSOR_DATE_POSITION);
            } else if ($lastTaskExecution) {
                $this->lastExecutionTime = $lastTaskExecution->format('c');
            } else {
                $now = new \DateTime('now');
                $minutes10Ago = new \DateTime('now');
                $minutes10Ago->setTimestamp($now->getTimestamp()-(60*10));
                $this->lastExecutionTime = $minutes10Ago->format('c');
            }

            $this->run();
        } catch (\Throwable $e) {
            $this->markTaskFailed($task);
            $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportCustomerUpdates");
            return;
        }

        $this->rescheduleTask($task, $taskEntity);
    }

    /**
     * @return iterable
     */
    public static function getHandledMessages(): iterable
    {
        return [ExportCustomerUpdatesTask::class];
    }

    /**
     * @return void
     */
    public function run(): void
    {
        try {
            $customerData = [];
            $customerUpdateFields = [
                "external_id",
                "customer_id",
                "email",
                "firstname",
                "lastname",
                "salutation",
                "salutation_readable",
                "store_id",
                "locale",
                "created_at",
                "updated_at",
                "geo_location",
                "guest"
            ];

            $allSalesChannels = $this->getAllSalesChannels();

            $customers = $this->getCustomers();
            if ($customers && count($customers)) {
                $limitReached = count($customers)>=self::CUSTOMER_LIMIT_PER_REQUEST;
                $lastUpdatedDate = null;
                foreach ($customers as $customer) {
                    if (!$customer || !$customer->getId() || !$customer->getEmail()) {
                        continue;
                    }
                    $salesChannelId = $customer->getSalesChannelId();
                    $lastUpdatedDate = $customer->getUpdatedAt();
                    $locale = $this->getLocale($customer);
                    $country = $this->getCountry($customer);

                    $enrichedCustomer = [
                        'external_id'               => $customer->getCustomerNumber() ?? $customer->getId(),
                        'customer_id'               => $customer->getId(),
                        'email'                     => strtolower($customer->getEmail()),
                        'firstname'                 => $customer->getFirstname(),
                        'lastname'                  => $customer->getLastname(),
                        'salutation'                => $customer->getSalutation()?->getSalutationKey(),
                        'salutation_readable'       => $customer->getSalutation()?->getLetterName(),
                        'store_id'                  => $salesChannelId,
                        'locale'                    => $locale?->getCode(),
                        'created_at'                => $customer->getCreatedAt()->format('c'),
                        'updated_at'                => $customer->getUpdatedAt()->format('c'),
                        'geo_location'              => [
                            'country'                   => $country?->getName(),
                        ],
                        'guest'                     => $customer->getGuest()
                    ];

                    // add JUNE collectID to customerData to match contacts based on ID rather then email
                    if ($customer?->getId()) {
                        $juneCustomer = $this->getJuneCustomerData($customer->getId(), $salesChannelId);
                        $juneCustomerCollectId = $juneCustomer?->getVars()['juneId'];
                        if ($juneCustomerCollectId) {
                            $enrichedCustomer['id'] = $juneCustomerCollectId;
                        }
                    }

                    // add more information to customer if available
                    if ($customer?->getCustomFields()) {
                        foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                            if (!isset($enrichedCustomer[$fieldName])) {
                                $transformedFieldName = ($enrichedCustomer['guest'] ? 'guest_' : '') . $fieldName;
                                $enrichedCustomer[$transformedFieldName] = $fieldValue;

                                if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                    $customerUpdateFields[] = $transformedFieldName;
                                }
                            }
                        }
                    }
                    if ($customer?->getBirthday()) {
                        $enrichedCustomer['birthday'] = $customer->getBirthday()->format('c');
                        if (!in_array('birthday', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'birthday';
                        }
                    }
                    if ($customer?->getLastLogin()) {
                        $enrichedCustomer['last_login'] = $customer->getLastLogin()->format('c');
                        if (!in_array('last_login', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_login';
                        }
                    }
                    if ($customer?->getLastOrderDate()) {
                        $enrichedCustomer['last_order_date'] = $customer->getLastOrderDate()->format('c');
                        if (!in_array('last_order_date', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_order_date';
                        }
                    }
                    if ($customer?->getOrderCount()) {
                        $enrichedCustomer['order_count'] = $customer->getOrderCount();
                        if (!in_array('order_count', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_count';
                        }
                    }
                    if ($customer?->getOrderTotalAmount()) {
                        $enrichedCustomer['order_total_amount'] = $customer->getOrderTotalAmount();
                        if (!in_array('order_total_amount', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_total_amount';
                        }
                    }

                    if ($customer?->getCampaignCode()) {
                        $enrichedCustomer['campaign_code'] = $customer->getCampaignCode();
                        if (!in_array('campaign_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'campaign_code';
                        }
                    }

                    if ($customer?->getAffiliateCode()) {
                        $enrichedCustomer['affiliate_code'] = $customer->getAffiliateCode();
                        if (!in_array('affiliate_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'affiliate_code';
                        }
                    }

                    if ($customer?->getTagIds()) {
                        $enrichedCustomer['_tags'] = [];
                        foreach($customer->getTagIds() AS $tagId) {
                            if ($tag = $this->getTag($tagId)) {
                                $enrichedCustomer['_tags'][] = $tag;
                            }
                        }
                        if (!in_array('_tags', $customerUpdateFields)) {
                            $customerUpdateFields[] = '_tags';
                        }
                    }

                    if ($customer?->getGroup()?->getName()) {
                        $enrichedCustomer['customer_group'] = $customer->getGroup()->getName();
                        if (!in_array('customer_group', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'customer_group';
                        }
                    }

                    // if customer is assigned to single salesChannel, we define it for Export
                    if ($salesChannelId) {
                        if (!isset($customerData[$salesChannelId])) {
                            $customerData[$salesChannelId] = [];
                        }

                        $this->enrichSubscriberInformation($customer, $enrichedCustomer, $customerUpdateFields, $salesChannelId);
                        $customerData[$salesChannelId][$customer->getEmail()] = $enrichedCustomer;
                    }
                    // if no salesChannels are assigned we loop through all channels
                    else if (!empty($allSalesChannels)) {
                        foreach ($allSalesChannels AS $singleSalesChannel) {
                            if (!isset($customerData[$singleSalesChannel->getId()])) {
                                $customerData[$singleSalesChannel->getId()] = [];
                            }

                            $enrichedCustomer['store_id'] = $singleSalesChannel->getId();
                            $this->enrichSubscriberInformation($customer, $enrichedCustomer, $customerUpdateFields, $singleSalesChannel->getId());
                            $customerData[$singleSalesChannel->getId()][$customer->getEmail()] = $enrichedCustomer;
                        }
                    }
                }

                // send collected data to June
                if (!empty($customerData)) {
                    foreach ($customerData AS $salesChannelId => $customers) {
                        $commerceCollectToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_TOKEN, $salesChannelId);
                        $commerceCollectAuth = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_AUTH, $salesChannelId);

                        if($this->configService->get(JuneCommerceSubscriber::JUNE_TRANSFER_CUSTOMERS, $salesChannelId) && !empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
                            // recreate comparison date for loading users to log it
                            $formatedUtcNoOffset = new \DateTime($this->lastExecutionTime);
                            $formatedUtcNoOffset = $formatedUtcNoOffset->format('Y-m-d\TH:i:s');

                            $this->loggerHelper->logInfo("Scheduled customer update export started for salesChannel '" . $salesChannelId . "' containing " . count($customers) . " customers created/updated after $formatedUtcNoOffset.", "exportCustomerUpdates");

                            $requestData = [
                                "action" => "customer_update",
                                "json" => array_values($customers),
                                "activate_trigger" => true,
                                "default_subscribe" => false,
                                "force_subscribe" => false,
                                "update_if_exists" => true,
                                "overwrite_tags" => true,
                                "update_fields" => $customerUpdateFields
                            ];

                            $importLoggerResponse = $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth, $this->maxExecutionTime);
                            if ($importLoggerResponse?->getBody()) {
                                $importLoggerToken = $importLoggerResponse->getBody()->getContents();
                                $this->loggerHelper->logInfo("Scheduled customer update export logger token for last execution: ".var_export($importLoggerToken, true), "exportCustomerUpdates");
                            }
                        } else {
                            if (empty($commerceCollectToken)) {
                                $this->loggerHelper->logInfo("Scheduled customer update export for salesChannel '" . $salesChannelId . "' could not run because CommerceCollectToken is not configured in the JUNE extension configuration.", "exportCustomerUpdates");
                            } elseif (empty($commerceCollectAuth)) {
                                $this->loggerHelper->logInfo("Scheduled customer update export for salesChannel '" . $salesChannelId . "' could not run because CommerceCollectAUTH is not configured in the JUNE extension configuration.", "exportCustomerUpdates");
                            }
                        }
                    }
                }

                if ($limitReached) {
                    // increment pagination for next run
                    $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) + 1);
                    $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $this->lastExecutionTime);
                } else {
                    // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                    $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                    $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
                }
            }
            // no customers found in this execution
            else {
                // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
            }
        } catch (\Exception $e) {
            // set last execution for retry
            $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $this->lastExecutionTime);

            $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportCustomerUpdates");
        }
    }

    /**
     * @param CustomerEntity $user
     * @param $enrichedCustomer
     * @param $customerUpdateFields
     * @param $salesChannelId
     * @return void
     */
    private function enrichSubscriberInformation(CustomerEntity $user, &$enrichedCustomer, &$customerUpdateFields, $salesChannelId) {
        $subscriber = $this->getNewsletterSubscriber($user->getEmail(), $salesChannelId);

        if ($subscriber) {
            // subscription
            $enrichedCustomer['_unsubscribed'] = !($subscriber->getStatus() === "direct" || $subscriber->getStatus() === "optIn");
            if (!in_array('_unsubscribed', $customerUpdateFields)) {
                $customerUpdateFields[] = '_unsubscribed';
            }

            $enrichedCustomer['intent'] = $subscriber->getStatus();
            if (!in_array('intent', $customerUpdateFields)) {
                $customerUpdateFields[] = 'intent';
            }

            if ($subscriber->getConfirmedAt()) {
                $enrichedCustomer['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                if (!in_array('subscription_confirmed_at', $customerUpdateFields)) {
                    $customerUpdateFields[] = 'subscription_confirmed_at';
                }
            }

            // location information
            if ($subscriber->getCity()) {
                $enrichedCustomer['city'] = $subscriber->getCity();
                $enrichedCustomer['geo_location']['city'] = $subscriber->getCity();
                if (!in_array('city', $customerUpdateFields)) {
                    $customerUpdateFields[] = 'city';
                }
            }
            if ($subscriber->getZipCode()) {
                $enrichedCustomer['zip_code'] = $subscriber->getZipCode();
                if (!in_array('zip_code', $customerUpdateFields)) {
                    $customerUpdateFields[] = 'zip_code';
                }
            }

            // tags
            if ($subscriber->getTags()) {
                if (!isset($enrichedCustomer['_tags'])) {
                    $enrichedCustomer['_tags'] = [];
                }
                foreach ($subscriber->getTags() as $tag) {
                    if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                        $enrichedCustomer['_tags'][] = $tag->getName();
                    }
                }
            }
        }
    }

    /**
     * @return array<SalesChannelEntity>
     */
    private function getAllSalesChannels(): array {
        $channelActiveFilter = new Criteria();
        $channelActiveFilter->addFilter(new EqualsFilter("active", true));
        return $this->salesChannelRepository->search($channelActiveFilter, Context::createDefaultContext())->getElements();
    }

    /**
     * @return array<CustomerEntity>
     */
    private function getCustomers(): array {
        if (!$this->lastExecutionTime) {
            $now = new \DateTime('now');
            $minutes10Ago = new \DateTime('now');
            $minutes10Ago->setTimestamp($now->getTimestamp()-(60*10));
            $this->lastExecutionTime = $minutes10Ago->format('c');
        }

        $lastUpdatedUserPage = $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) ?? 0;

        // format updatedAt date and remove offset because Shopware adds it themselves
        $formatedUtcNoOffset = new \DateTime($this->lastExecutionTime);
        $formatedUtcNoOffset = $formatedUtcNoOffset->format('Y-m-d\TH:i:s');

        $customerCriteria = new Criteria();
        $customerCriteria->addFilter(new OrFilter(
            [
                new RangeFilter('createdAt',
                    [
                        RangeFilter::GTE => $formatedUtcNoOffset
                    ]
                ),
                new RangeFilter('updatedAt',
                    [
                        RangeFilter::GTE => $formatedUtcNoOffset
                    ]
                )
            ])
        );
        $customerCriteria->addSorting(new FieldSorting('updatedAt', FieldSorting::ASCENDING));
        $customerCriteria->setLimit(self::CUSTOMER_LIMIT_PER_REQUEST);
        if ($lastUpdatedUserPage) {
            $customerCriteria->setOffset($lastUpdatedUserPage * self::CUSTOMER_LIMIT_PER_REQUEST);
        }
        $customerCriteria->addAssociation('tags');
        $customerCriteria->addAssociation('salutation');
        $customerCriteria->addAssociation('group');
        $customerCriteria->addAssociation('juneCustomer');

        // $this->loggerHelper->logInfo("Scheduled customer query: Offset:".$lastUpdatedUserPage * self::CUSTOMER_LIMIT_PER_REQUEST ." - Limit:".self::CUSTOMER_LIMIT_PER_REQUEST ." - UpdatedAt>=".$formatedUtcNoOffset, "exportCustomerUpdates");

        return $this->customerRepository->search($customerCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * @param string $email
     * @param string $salesChannelId
     * @return NewsletterRecipientEntity|null
     */
    private function getNewsletterSubscriber(string $email, string $salesChannelId): ?NewsletterRecipientEntity {
        if ($email) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $email));
            $itemCriteria->addFilter(new EqualsFilter('salesChannelId', $salesChannelId));
            $itemCriteria->addAssociation('tags');
            return $this->newsletterRecipientRepository->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }

    /**
     * @param CustomerEntity $customer
     * @return LocaleEntity|null
     */
    private function getLocale(CustomerEntity $customer): ?LocaleEntity {
        $locale = null;

        if ($customer->getLanguageId()) {
            $languageCriteria = new Criteria();
            $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $customer->getLanguageId()));
            $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

            if ($language->getLocaleId()) {
                $localeCriteria = new Criteria();
                $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
                $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();
            }
        }

        return $locale;
    }

    /**
     * @param CustomerEntity $customer
     * @return CountryEntity|null
     */
    private function getCountry(CustomerEntity $customer): ?CountryEntity {
        $country = null;

        if ($customer->getSalesChannelId()) {
            $salesChannelCriteria = new Criteria();
            $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $customer->getSalesChannelId()));
            $salesChannel = $this->salesChannelRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

            if ($salesChannel->getCountryId()) {
                $countryCriteria = new Criteria();
                $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
                $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();
            }
        }

        return $country;
    }

    /**
     * @param string $tagId
     * @return string
     */
    private function getTag(string $tagId): string {
        if (!in_array($tagId, $this->loadedSubscriberTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedSubscriberTags[$tagId] = $tag->getName();
        }

        return $this->loadedSubscriberTags[$tagId];
    }

    /**
     * @param $customerId
     * @param $salesChannelId
     * @return \Shopware\Core\Framework\DataAbstractionLayer\Entity|null
     */
    private function getJuneCustomerData($customerId, $salesChannelId) {
        $searchCriteria = new Criteria();
        $searchCriteria->addFilter(new EqualsFilter('customerId', $customerId));
        $searchCriteria->addFilter(new EqualsFilter('salesChannelId', $salesChannelId));

        return $this->juneCustomerRepository->search($searchCriteria, Context::createDefaultContext())->first();
    }
}
