<?php

namespace June\Service\ScheduledTask;

use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Checkout\Order\Aggregate\OrderCustomer\OrderCustomerEntity;
use Shopware\Core\Checkout\Order\OrderEntity;
use Shopware\Core\Checkout\Order\Aggregate\OrderLineItem\OrderLineItemEntity;
use Shopware\Core\Content\Product\ProductEntity;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\OrFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Sorting\FieldSorting;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTask;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskDefinition;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskEntity;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskHandler;
use Shopware\Core\System\SalesChannel\Context\AbstractSalesChannelContextFactory;
use Shopware\Core\System\SalesChannel\Entity\SalesChannelRepository;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use June\Subscriber\JuneCommerceSubscriber;

use Shopware\Core\System\Locale\LocaleEntity;
use Shopware\Core\System\Country\CountryEntity;
use Shopware\Core\Content\Product\Aggregate\ProductMedia\ProductMediaEntity;
use Shopware\Core\System\Tag\TagEntity;

#[AsMessageHandler(handles: ExportNewOrdersTask::class)]
class ExportNewOrdersTaskHandler extends ScheduledTaskHandler {

    const ORDER_LIMIT_PER_REQUEST = 50;
    const CONF_CURSOR_DATE_POSITION = 'June.config.last_order_updated_at_cursor';
    const CONF_CURSOR_PAGINATION_POSITION = 'June.config.last_order_pagination_cursor';

    private EntityRepository $orderRepository;

    private EntityRepository $pluginRepository;

    private EntityRepository $orderLineItemsRepository;

    private EntityRepository $productRepository;

    private EntityRepository $productMediaRepository;

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $newsletterRecipientEntity;

    private EntityRepository $customerEntity;

    private EntityRepository $customerTagsRepository;

    private SalesChannelRepository $categoryRepository;

    private AbstractSalesChannelContextFactory $salesChannelContextFactory;

    private EntityRepository $salesRepository;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private JuneCommerceSubscriber $juneCommerceSubscriber;

    private ?string $lastExecutionTime = null;

    private int $maxExecutionTime = 45;

    private array $loadedCustomerTags = [];

    public function __construct
    (
        EntityRepository $scheduledTaskRepository,
        EntityRepository $pluginRepository,
        EntityRepository $orderRepository,
        EntityRepository $orderLineItemsRepository,
        EntityRepository $productRepository,
        EntityRepository $productMediaRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $newsletterRecipientEntity,
        EntityRepository $customerEntity,
        EntityRepository $customerTagsRepository,
        SalesChannelRepository $categoryRepository,
        AbstractSalesChannelContextFactory $salesChannelContextFactory,
        EntityRepository $salesRepository,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient,
        JuneCommerceSubscriber $juneCommerceSubscriber
    )
    {
        parent::__construct($scheduledTaskRepository);
        $this->orderRepository = $orderRepository;
        $this->pluginRepository = $pluginRepository;
        $this->orderLineItemsRepository = $orderLineItemsRepository;
        $this->productRepository = $productRepository;
        $this->productMediaRepository = $productMediaRepository;
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->newsletterRecipientEntity = $newsletterRecipientEntity;
        $this->customerEntity = $customerEntity;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->categoryRepository = $categoryRepository;
        $this->salesChannelContextFactory = $salesChannelContextFactory;
        $this->salesRepository = $salesRepository;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;
        $this->juneCommerceSubscriber = $juneCommerceSubscriber;

        $serverMaxExecutionTime = ini_get('max_execution_time');
        if ($serverMaxExecutionTime && $serverMaxExecutionTime > 0) {
            $this->maxExecutionTime = $serverMaxExecutionTime - 5;

            if ($this->maxExecutionTime <= 0) {
                $this->maxExecutionTime = $serverMaxExecutionTime * 1;
            }
        }

        $this->lastExecutionTime = null;
    }

    /**
     * @param ScheduledTask $task
     * @return void
     */
    public function __invoke(ScheduledTask $task): void
    {
        $taskId = $task->getTaskId();
        if ($taskId === null) {
            // run task independent of the schedule
            $this->run();
            return;
        }

        /** @var ScheduledTaskEntity|null $taskEntity */
        $taskEntity = $this->scheduledTaskRepository
            ->search(new Criteria([$taskId]), Context::createDefaultContext())
            ->get($taskId);

        if ($taskEntity === null || !$taskEntity->isExecutionAllowed()) {
            return;
        }

        $this->markTaskRunning($task);
        try {
            $lastTaskExecution = $taskEntity->getLastExecutionTime();

            if ($this->configService->get(self::CONF_CURSOR_DATE_POSITION)) {
                $this->lastExecutionTime = $this->configService->get(self::CONF_CURSOR_DATE_POSITION);
            } else if ($lastTaskExecution) {
                $this->lastExecutionTime = $lastTaskExecution->format('c');
            } else {
                $now = new \DateTime();
                $minutes10Ago = new \DateTime();
                $minutes10Ago->setTimestamp($now->getTimestamp()-(60*10));
                $this->lastExecutionTime = $minutes10Ago->format('c');
            }

            $this->run();
        } catch (\Throwable $e) {
            $this->markTaskFailed($task);
            $this->loggerHelper->logError($e->getMessage(), "exportOrders");
            return;
        }

        $this->rescheduleTask($task, $taskEntity);
    }

    /**
     * @return iterable
     */
    public static function getHandledMessages(): iterable
    {
        return [ExportNewOrdersTask::class];
    }

    /**
     * @return void
     */
    public function run(): void
    {
        try {
            $orderData = [];
            $orders = $this->getOrders();
            if ($orders && count($orders)) {
                $limitReached = count($orders)>=self::ORDER_LIMIT_PER_REQUEST;
                $lastUpdatedDate = null;
                foreach ($orders as $order) {
                    $lastUpdatedDate = $order->getUpdatedAt() ?? $order->getCreatedAt();

                    if (!$order || !$order->getOrderNumber()) {
                        continue;
                    }

                    $salesChannelId = $order->getSalesChannelId();
                    $commerceToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_TOKEN, $salesChannelId);
                    if(empty($commerceToken) || !$this->configService->get(JuneCommerceSubscriber::JUNE_TRANSFER_ORDER, $salesChannelId)) {
                        if (empty($commerceToken)) {
                            $this->loggerHelper->logInfo("Scheduled order export for salesChannel '".$salesChannelId."' could not run because CommerceToken is not configured in the JUNE extension configuration.", "exportOrders");
                        }
                        continue;
                    }

                    $voucher = 0;
                    $shopUrl = $this->getShopUrl($salesChannelId, $order->getLanguageId());
                    //$orderCustomer = ($order->getOrderCustomer() && $order->getOrderCustomer()->getCustomer() ? $order->getOrderCustomer()->getCustomer() : $order->getOrderCustomer());
                    $orderCustomer = $order->getOrderCustomer();
                    $customer = $orderCustomer->getCustomer() ?? $this->getCustomer($orderCustomer, $salesChannelId);

                    $products = [];
                    $lineItems = $this->getLineItems($order);

                    foreach ($lineItems as $item) {
                        // transform products and add voucher values to referenced variable
                        if ($product = $this->transformLineItem($item, $shopUrl, $salesChannelId, $voucher)) {
                            $products[] = $product;
                        }
                    }

                    // load locale
                    $locale = $this->getLocale($order);

                    // load country
                    $country = $this->getCountry($order);

                    // load id and email, prefers customer over order
                    $userId = $orderCustomer?->getCustomerNumber() ?? ($customer?->getId() ?? $orderCustomer?->getId());
                    $email = $customer?->getEmail() ?? ($orderCustomer?->getEmail() ?? '');
                    $email = strtolower($email);

                    $enrichedCustomer = [
                        'external_id' => $userId,
                        'firstname' => $orderCustomer?->getFirstname(),
                        'lastname' => $orderCustomer?->getLastname(),
                        'store_id' => $salesChannelId,
                        'locale' => $locale?->getCode(),
                        'geo_location' => [
                            'country' => $country?->getName(),
                        ],
                        'guest'         => !$customer || $customer->getGuest()
                    ];

                    if ($customer?->getGroup()?->getName()) {
                        $enrichedCustomer['customer_group'] = $customer->getGroup()->getName();
                    }

                    $enrichedOrder = [
                        'id' => $userId,
                        'email' => $email,
                        'customer' => $enrichedCustomer,
                        'order' => [
                            'id' => $order->getOrderNumber(),
                            'created_at' => $order->getOrderDateTime()->format('c'),
                            'total' => ($order->getPrice() ? $order->getPrice()->getTotalPrice() : null),
                            'shipping' => ($order->getShippingCosts() ? $order->getShippingCosts()->getTotalPrice() : null),
                            'voucher' => $voucher,
                            'products' => $products,

                            'currency' => $order->getCurrency()?->getShortName(),
                            'affiliate_code' => $order->getAffiliateCode(),
                            'campaign_id' => $order->getCampaignCode(),
                            'shipping_method' => $order->getDeliveries()?->getShippingMethods()?->first()?->getName(),
                            'payment_method' => $order->getTransactions()?->first()?->getPaymentMethod()?->getName(),
                        ]
                    ];

                    if ($order->getDeliveries()) {
                        foreach ($order->getDeliveries() AS $delivery) {
                            if ($delivery->getStateMachineState()->getTechnicalName() === 'shipped') {
                                $enrichedOrder['order']['fulfilled_at'] = $delivery->getUpdatedAt()->format('c');
                            }
                        }
                    }

                    if (!isset($orderData[$salesChannelId])) {
                        $orderData[$salesChannelId] = [];
                    }
                    $orderData[$salesChannelId][] = $enrichedOrder;

                    // set new updatedAt date for customer so all properties are exported on the next schedule run
                    if ($customer && $this->configService->get(JuneCommerceSubscriber::JUNE_TRANSFER_CUSTOMERS, $salesChannelId)) {
                        $this->juneCommerceSubscriber->createJuneCustomer($customer, $salesChannelId, $this->juneCommerceSubscriber->getLanguageCode($customer->getLanguageId()));
                    }
                }

                // send collected commerce data to June
                foreach ($orderData AS $salesChannelId => $orders) {
                    $commerceToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_TOKEN, $salesChannelId);
                    if ($commerceToken) {
                        $this->juneApiClient->sendData($orders, "https://engagement.juneapp.com/commerce/webhook/$commerceToken?import=true", null, $this->maxExecutionTime);
                    }
                }

                if ($limitReached) {
                    // increment pagination for next run
                    $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) + 1);
                    $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $this->lastExecutionTime);
                } else {
                    // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                    $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                    $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
                }
            }
            // no orders found for this run
            else {
                // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
            }
        } catch (\Exception $e) {
            // set last execution for retry
            $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $this->lastExecutionTime);

            $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportOrders");
        }
    }

    /**
     * @param OrderLineItemEntity $item
     * @param $shopUrl
     * @param $salesChannelId
     * @param $voucher
     * @return array|void
     */
    private function transformLineItem(OrderLineItemEntity $item, $shopUrl, $salesChannelId, &$voucher) {
        try {
            if ($item->getType() === "promotion") {
                $voucher = $voucher + round($item->getTotalPrice(), 2);
            }

            $product = $this->getProduct($item->getProductId());
            $productData = $item->getPayload();
            $productExternalIdentifier = !empty($productData['productNumber']) ? $productData['productNumber'] : $item->getId();
            $productCover = $this->getProductCover($item->getCoverId() ?? $product?->getCoverId());
            $productCoverUrl = ($productCover && $productCover->getMedia() && $productCover->getMedia()->getUrl()) ? $productCover->getMedia()->getUrl() : ($item?->getCover() ? $item?->getCover()?->getUrl() : ($product?->getCover() ? $product?->getCover()?->getMedia()?->getUrl() : null));

            $categories = [];
            if (!empty($productData['categoryIds'])) {
                foreach ($productData['categoryIds'] as $categoryId) {
                    if ($categoryId) {
                        $categoryCriteria = new Criteria();
                        $categoryCriteria = $categoryCriteria->addFilter(new EqualsFilter('id', $categoryId));
                        $category = $this->categoryRepository->search($categoryCriteria, $this->createSalesChannelContext($salesChannelId))->first();
                        if ($category?->getName()) {
                            $categories[] = $category->getName();
                        }
                    }
                }
            }

            // load variant definitions for cart items
            $productName = $item->getLabel();
            if (!empty($productData['options'])) {
                foreach ($productData['options'] as $optionItem) {
                    if (!empty($optionItem['option'])) {
                        $productName .= " | " . $optionItem['option'];
                    }
                }
            }

            $productData = [
                'id' => $productExternalIdentifier,
                'name' => $productName,
                'qty' => $item->getQuantity(),
                'price' => $item->getPrice()?->getUnitPrice(),
                'category' => $categories
            ];

            if ($productCoverUrl) {
                $productData['image'] = $productCoverUrl;
            }

            if ($shopUrl && $item->getProductId()) {
                $productData['link'] = $shopUrl . "/detail/" . $item->getProductId();
            }

            return $productData;
        } catch (\Exception $e) {
            $this->loggerHelper->logError($e->getMessage(), "exportTransformLineItems");
        }
    }

    /**
     * @param string $salesChannelId
     * @return SalesChannelContext
     */
    private function createSalesChannelContext(string $salesChannelId): SalesChannelContext
    {
        return $this->salesChannelContextFactory->create('', $salesChannelId, []);
    }

    /**
     * @param $salesChannelId
     * @param $languageId
     * @return string|null
     */
    private function getShopUrl($salesChannelId, $languageId): ?string {
        $salesChannelUrls = [];
        $salesCriteria = new Criteria();
        $salesCriteria = $salesCriteria->addAssociation('domains');
        $salesChannelIds = $this->salesRepository->search($salesCriteria, Context::createDefaultContext());
        foreach($salesChannelIds->getEntities()->getElements() as $key => $salesChannel){
            foreach($salesChannel->getDomains() as $element) {
                $salesChannelUrls[] = $element;
            }
        }
        foreach ($salesChannelUrls as $url) {
            if($url->getSalesChannelId() === $salesChannelId && $url->getLanguageId() === $languageId) {
                return $url->getUrl();
            }
        }
        return null;
    }

    /**
     * @return array<OrderEntity>
     */
    private function getOrders(): array {
        $lastUpdatedOrderPage = $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) ?? 0;

        // format updatedAt date and remove offset because Shopware adds it themselves
        $formatedUtcNoOffset = new \DateTime($this->lastExecutionTime);
        $formatedUtcNoOffset = $formatedUtcNoOffset->format('Y-m-d\TH:i:s');

        $orderCriteria = new Criteria();
        $orderCriteria->addFilter(new OrFilter(
            [
                new RangeFilter('createdAt',
                    [
                        RangeFilter::GTE => $formatedUtcNoOffset
                    ]
                ),
                new RangeFilter('updatedAt',
                    [
                        RangeFilter::GTE => $formatedUtcNoOffset
                    ]
                )
            ]
        ));

        $orderCriteria->addSorting(new FieldSorting('updatedAt', FieldSorting::ASCENDING));
        $orderCriteria->setLimit(self::ORDER_LIMIT_PER_REQUEST);
        if ($lastUpdatedOrderPage) {
            $orderCriteria->setOffset($lastUpdatedOrderPage * self::ORDER_LIMIT_PER_REQUEST);
        }
        //@TODO: rework orders to directly use lineItem association
        //$orderCriteria->addAssociation('lineItems');
        $orderCriteria->addAssociation('orderCustomer');
        $orderCriteria->addAssociation('currency');
        $orderCriteria->addAssociation('billingAddress.country');
        $orderCriteria->addAssociation('deliveries.shippingMethod.name');
        $orderCriteria->addAssociation('transactions.paymentMethod');

        return $this->orderRepository->search($orderCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * @param OrderEntity $order
     * @return array<OrderLineItemEntity>
     */
    private function getLineItems(OrderEntity $order): array {
        $itemCriteria = new Criteria();
        $itemCriteria->addFilter(new EqualsFilter("orderId", $order->getId()));
        $itemCriteria->addAssociation('cover');
        return $this->orderLineItemsRepository->search($itemCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * @param $productId
     * @return ProductEntity|null
     */
    private function getProduct($productId): ?ProductEntity {
        if ($productId) {
            $itemCriteria = new Criteria([$productId]);
            $itemCriteria->addAssociation('cover');
            return $this->productRepository->search(new $itemCriteria, Context::createDefaultContext())->first();
        }

        return null;
    }

    /**
     * @param $mediaId
     * @return ProductMediaEntity|null
     */
    private function getProductCover($mediaId): ?ProductMediaEntity {
        return ($mediaId ? $this->productMediaRepository->search(new Criteria([$mediaId]), Context::createDefaultContext())->first() : null);
    }

    /**
     * @param OrderEntity $order
     * @return LocaleEntity|null
     */
    private function getLocale(OrderEntity $order): ?LocaleEntity {
        $locale = null;

        if ($order->getLanguageId()) {
            $languageCriteria = new Criteria();
            $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $order->getLanguageId()));
            $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

            if ($language->getLocaleId()) {
                $localeCriteria = new Criteria();
                $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
                $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();
            }
        }

        return $locale;
    }

    /**
     * @param OrderEntity $order
     * @return CountryEntity|null
     */
    private function getCountry(OrderEntity $order): ?CountryEntity {
        $country = null;

        if ($order->getSalesChannelId()) {
            $salesChannelCriteria = new Criteria();
            $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $order->getSalesChannelId()));
            $salesChannel = $this->salesRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

            if ($salesChannel->getCountryId()) {
                $countryCriteria = new Criteria();
                $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
                $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();
            }
        }

        return $country;
    }

    /**
     * @param OrderCustomerEntity $orderCustomer
     * @return CustomerEntity|null
     */
    private function getCustomer(OrderCustomerEntity $orderCustomer, string $salesChannelId): ?CustomerEntity {
        if ($orderCustomer) {
            if ($orderCustomer->getCustomerId()) {
                return $this->customerEntity->search(new Criteria([$orderCustomer->getCustomerId()]), Context::createDefaultContext())->first();
            } else if ($orderCustomer->getCustomerNumber()) {
                $itemCriteria = new Criteria();
                $itemCriteria->addFilter(new EqualsFilter("customerNumber", $orderCustomer->getCustomerNumber()));
                $itemCriteria->addFilter(new EqualsFilter("salesChannelId", $salesChannelId));
                $itemCriteria->addAssociation('tags');
                $itemCriteria->addAssociation('group');
                $itemCriteria->addAssociation('juneCustomer');
                return $this->customerEntity->search($itemCriteria, Context::createDefaultContext())->first();
            }
        }
        return null;
    }

    /**
     * @param string $tagId
     * @return string
     */
    private function getTag(string $tagId) :string {
        if (!in_array($tagId, $this->loadedCustomerTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedCustomerTags[$tagId] = $tag->getName();
        }

        return $this->loadedCustomerTags[$tagId];
    }

    /**
     * @param OrderCustomerEntity|null $orderCustomer
     * @return NewsletterRecipientEntity|null
     */
    private function getNewsletterSubscriber(?OrderCustomerEntity $orderCustomer): ?NewsletterRecipientEntity {
        if ($orderCustomer && $orderCustomer->getEmail()) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $orderCustomer->getEmail()));
            $itemCriteria->addAssociation('tags');
            return $this->newsletterRecipientEntity->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }
}
