<?php

namespace June\Service\ScheduledTask;

use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use Psr\Log\LoggerInterface;
use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Checkout\Order\Aggregate\OrderCustomer\OrderCustomerEntity;
use Shopware\Core\Checkout\Order\OrderEntity;
use Shopware\Core\Checkout\Order\Aggregate\OrderLineItem\OrderLineItemEntity;
use Shopware\Core\Content\Product\ProductEntity;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsAnyFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\MultiFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NorFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NotFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTask;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskDefinition;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskEntity;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskHandler;
use Shopware\Core\System\SalesChannel\Context\AbstractSalesChannelContextFactory;
use Shopware\Core\System\SalesChannel\Entity\SalesChannelRepository;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use June\Subscriber\JuneCommerceSubscriber;

use Shopware\Core\System\Locale\LocaleEntity;
use Shopware\Core\System\Country\CountryEntity;
use Shopware\Core\Content\Product\Aggregate\ProductMedia\ProductMediaEntity;
use Shopware\Core\System\Tag\TagEntity;
use Symfony\Component\Messenger\Attribute\AsMessageHandler;

#[AsMessageHandler(handles: ExportOrdersTask::class)]
class ExportOrdersTaskHandler extends ScheduledTaskHandler {

    const ORDER_LIMIT_PER_REQUEST = 25;

    private EntityRepository $orderRepository;

    private EntityRepository $pluginRepository;

    private EntityRepository $orderLineItemsRepository;

    private EntityRepository $productRepository;

    private EntityRepository $productMediaRepository;

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $newsletterRecipientEntity;

    private EntityRepository $customerEntity;

    private EntityRepository $customerTagsRepository;

    private SalesChannelRepository $categoryRepository;

    private AbstractSalesChannelContextFactory $salesChannelContextFactory;

    private EntityRepository $salesRepository;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private array $loadedCustomerTags = [];

    private int $maxExecutionTime = 25;

    public function __construct
    (
        EntityRepository $scheduledTaskRepository,
        EntityRepository $pluginRepository,
        EntityRepository $orderRepository,
        EntityRepository $orderLineItemsRepository,
        EntityRepository $productRepository,
        EntityRepository $productMediaRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $newsletterRecipientEntity,
        EntityRepository $customerEntity,
        EntityRepository $customerTagsRepository,
        SalesChannelRepository $categoryRepository,
        AbstractSalesChannelContextFactory $salesChannelContextFactory,
        EntityRepository $salesRepository,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient,
        LoggerInterface $logger,
    )
    {
        parent::__construct($scheduledTaskRepository, $logger);
        $this->orderRepository = $orderRepository;
        $this->pluginRepository = $pluginRepository;
        $this->orderLineItemsRepository = $orderLineItemsRepository;
        $this->productRepository = $productRepository;
        $this->productMediaRepository = $productMediaRepository;
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->newsletterRecipientEntity = $newsletterRecipientEntity;
        $this->customerEntity = $customerEntity;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->categoryRepository = $categoryRepository;
        $this->salesChannelContextFactory = $salesChannelContextFactory;
        $this->salesRepository = $salesRepository;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;

        $serverMaxExecutionTime = ini_get('max_execution_time');
        if ($serverMaxExecutionTime && $serverMaxExecutionTime > 0) {
            $this->maxExecutionTime = $serverMaxExecutionTime - 5;
        }
        if ($this->maxExecutionTime <= 0) {
            $this->maxExecutionTime = $serverMaxExecutionTime * 1;
        }
    }

    /**
     * @param ScheduledTask $task
     * @return void
     */
    public function __invoke(ScheduledTask $task): void
    {
        $taskId = $task->getTaskId();
        if ($taskId === null) {
            // run task independent of the schedule
            $this->run();
            return;
        }

        /** @var ScheduledTaskEntity|null $taskEntity */
        $taskEntity = $this->scheduledTaskRepository
            ->search(new Criteria([$taskId]), Context::createDefaultContext())
            ->get($taskId);

        if ($taskEntity === null || !$taskEntity->isExecutionAllowed()) {
            //@DEV
            $this->loggerHelper->logInfo("Order export task status: ". $taskEntity->getStatus() ." not executed", "exportOrders");

            return;
        } else {
            //@DEV
            $this->loggerHelper->logInfo("Order export task status: ". $taskEntity->getStatus() ." is executed", "exportOrders");
        }

        if($this->configService->get('June.config.startExport') === true) {
            $this->markTaskRunning($task);
            try {
                $this->run();
            } catch (\Throwable $e) {
                $this->markTaskFailed($task);
                $this->loggerHelper->logError($e->getMessage(), "exportOrders");
                return;
            }

            $availableOrders = $this->getAvailableOrdersCount();
            if ($availableOrders > 0) {
                $this->rescheduleTask($task, $taskEntity);
            } else {
                $this->loggerHelper->logInfo("Order export finished for a total of ".$this->getTotalOrdersCount()." orders", "exportOrders");
                $this->scheduledTaskRepository->update([
                    [
                        'id' => $task->getTaskId(),
                        'status' => ScheduledTaskDefinition::STATUS_INACTIVE,
                    ],
                ], Context::createDefaultContext());
            }
        } else {
            $this->loggerHelper->logInfo("Order export inactive.", "exportOrders");
            $this->scheduledTaskRepository->update([
                [
                    'id' => $task->getTaskId(),
                    'status' => ScheduledTaskDefinition::STATUS_INACTIVE,
                ],
            ], Context::createDefaultContext());
        }
    }

    /**
     * @return iterable
     */
    public static function getHandledMessages(): iterable
    {
        return [ExportOrdersTask::class];
    }

    /**
     * @return void
     */
    public function run(): void
    {
        if($this->configService->get('June.config.startExport') === true) {
            try {
                $orderData = [];
                $scheduleOrderOffset = $this->configService->get('June.config.orderOffset');
                $orders = $this->getOrders();
                if ($orders) {
                    $this->loggerHelper->logInfo("Scheduled order export started for order " . $scheduleOrderOffset . "-" . ($scheduleOrderOffset + self::ORDER_LIMIT_PER_REQUEST), "exportOrders");

                    foreach ($orders as $order) {
                        if (!$order || !$order->getOrderNumber()) {
                            continue;
                        }

                        //$this->loggerHelper->logInfo("Exporting order ".$order->getOrderNumber()." ...", "exportOrders");

                        $salesChannelId = $order->getSalesChannelId();

                        $voucher = 0;
                        $shopUrl = $this->getShopUrl($salesChannelId, $order->getLanguageId());
                        //$orderCustomer = ($order->getOrderCustomer() && $order->getOrderCustomer()->getCustomer() ? $order->getOrderCustomer()->getCustomer() : $order->getOrderCustomer());
                        $orderCustomer = $order->getOrderCustomer();
                        $customer = $orderCustomer->getCustomer() ?? $this->getCustomer($orderCustomer, $salesChannelId);

                        $products = [];
                        $lineItems = $this->getLineItems($order);

                        foreach ($lineItems as $item) {
                            // transform products and add voucher values to referenced variable
                            if ($product = $this->transformLineItem($item, $shopUrl, $salesChannelId, $voucher)) {
                                $products[] = $product;
                            }
                        }

                        // load locale
                        $locale = $this->getLocale($order);

                        // load country
                        $country = $this->getCountry($order);

                        $userId = $orderCustomer?->getCustomerNumber() ?? ($customer?->getId() ?? $orderCustomer?->getId());
                        $email = $customer?->getEmail() ?? ($orderCustomer?->getEmail() ?? '');
                        $email = strtolower($email);

                        $enrichedCustomer = [
                            'external_id' => $userId,
                            'firstname' => $customer?->getFirstName() ?? $orderCustomer?->getFirstname(),
                            'lastname' => $customer?->getLastName() ?? $orderCustomer?->getLastname(),
                            'store_id' => $salesChannelId,
                            'locale' => $locale?->getCode(),
                            'geo_location' => [
                                'country' => $country?->getName(),
                            ],
                            'guest'         => !$customer || $customer->getGuest()
                        ];

                        /*
                        if ($customer?->getGroup()?->getName()) {
                            $enrichedCustomer['customer_group'] = $customer->getGroup()->getName();
                        }
                        */

                        $enrichedOrder = [
                            'id' => $userId,
                            'email' => $email,
                            'customer' => $enrichedCustomer,
                            'order' => [
                                'id' => $order->getOrderNumber(),
                                'created_at' => $order->getOrderDateTime()->format('c'),
                                'total' => ($order->getPrice() ? $order->getPrice()->getTotalPrice() : null),
                                'shipping' => ($order->getShippingCosts() ? $order->getShippingCosts()->getTotalPrice() : null),
                                'voucher' => $voucher,
                                'products' => $products,

                                'currency' => $order->getCurrency()?->getShortName(),
                                'affiliate_code' => $order->getAffiliateCode(),
                                'campaign_id' => $order->getCampaignCode(),
                                'shipping_method' => $order->getDeliveries()?->getShippingMethods()?->first()?->getName(),
                                'payment_method' => $order->getTransactions()?->first()?->getPaymentMethod()?->getName(),
                            ]
                        ];

                        if ($order->getDeliveries()) {
                            foreach ($order->getDeliveries() AS $delivery) {
                                if ($delivery->getStateMachineState()?->getTechnicalName() === 'shipped') {
                                    $enrichedOrder['order']['fulfilled_at'] = $delivery->getUpdatedAt()->format('c');
                                }
                            }
                        }

                        if (!isset($orderData[$salesChannelId])) {
                            $orderData[$salesChannelId] = [];
                        }
                        $orderData[$salesChannelId][] = $enrichedOrder;
                    }

                    foreach ($orderData AS $salesChannelId => $orders) {
                        $commerceToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_TOKEN, $salesChannelId);

                        if ($commerceToken) {
                            // send collected data to June
                            $this->juneApiClient->sendData($orders, "https://engagement.juneapp.com/commerce/webhook/$commerceToken?import=true", null, $this->maxExecutionTime);

                            //@DEV:
                            $this->loggerHelper->logInfo("Scheduled order export finished.", "exportOrders");
                        } else {
                            $this->loggerHelper->logInfo("Scheduled order export for salesChannel '".$salesChannelId."' could not run because CommerceToken is not configured in the JUNE extension configuration.", "exportOrders");
                        }
                    }
                }

                $this->configService->set('June.config.orderOffset', ($scheduleOrderOffset + self::ORDER_LIMIT_PER_REQUEST));
            } catch (\Exception $e) {
                $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportOrders");
            }
        }
    }

    /**
     * @param OrderLineItemEntity $item
     * @param $shopUrl
     * @param $salesChannelId
     * @param $voucher
     * @return array|void
     */
    private function transformLineItem(OrderLineItemEntity $item, $shopUrl, $salesChannelId, &$voucher) {
        try {
            if ($item->getType() === "promotion") {
                $voucher = $voucher + round($item->getTotalPrice(), 2);
            }

            $product = $this->getProduct($item->getProductId());
            $productData = $item->getPayload();
            $productExternalIdentifier = !empty($productData['productNumber']) ? $productData['productNumber'] : $item->getId();
            $productCover = $this->getProductCover($item->getCoverId() ?? $product?->getCoverId());
            $productCoverUrl = ($productCover && $productCover->getMedia() && $productCover->getMedia()->getUrl()) ? $productCover->getMedia()->getUrl() : ($item?->getCover() ? $item?->getCover()?->getUrl() : ($product?->getCover() ? $product?->getCover()?->getMedia()?->getUrl() : null));

            $categories = [];
            if (!empty($productData['categoryIds'])) {
                foreach ($productData['categoryIds'] as $categoryId) {
                    if ($categoryId) {
                        $categoryCriteria = new Criteria();
                        $categoryCriteria = $categoryCriteria->addFilter(new EqualsFilter('id', $categoryId));
                        $category = $this->categoryRepository->search($categoryCriteria, $this->createSalesChannelContext($salesChannelId))->first();
                        if ($category?->getName()) {
                            $categories[] = $category->getName();
                        }
                    }
                }
            }

            // load variant definitions for cart items
            $productName = $item->getLabel();
            if (!empty($productData['options'])) {
                foreach ($productData['options'] as $optionItem) {
                    if (!empty($optionItem['option'])) {
                        $productName .= " | " . $optionItem['option'];
                    }
                }
            }

            $productData = [
                'id' => $productExternalIdentifier,
                'name' => $productName,
                'qty' => $item->getQuantity(),
                'price' => $item->getPrice()?->getUnitPrice(),
                'category' => $categories
            ];

            if ($productCoverUrl) {
                $productData['image'] = $productCoverUrl;
            }

            if ($shopUrl && $item->getProductId()) {
                $productData['link'] = $shopUrl . "/detail/" . $item->getProductId();
            }

            return $productData;
        } catch (\Exception $e) {
            $this->loggerHelper->logError($e->getMessage(), "exportTransformLineItems");
        }
    }

    /**
     * @param string $salesChannelId
     * @return SalesChannelContext
     */
    private function createSalesChannelContext(string $salesChannelId): SalesChannelContext
    {
        return $this->salesChannelContextFactory->create('', $salesChannelId, []);
    }

    /**
     * @param $salesChannelId
     * @param $languageId
     * @return string|null
     */
    private function getShopUrl($salesChannelId, $languageId): ?string {
        $salesChannelUrls = [];
        $salesCriteria = new Criteria();
        $salesCriteria = $salesCriteria->addAssociation('domains');
        $salesChannelIds = $this->salesRepository->search($salesCriteria, Context::createDefaultContext());
        foreach($salesChannelIds->getEntities()->getElements() as $key => $salesChannel){
            foreach($salesChannel->getDomains() as $element) {
                $salesChannelUrls[] = $element;
            }
        }
        foreach ($salesChannelUrls as $url) {
            if($url->getSalesChannelId() === $salesChannelId && $url->getLanguageId() === $languageId) {
                return $url->getUrl();
            }
        }
        return null;
    }

    /**
     * @return array<OrderEntity>
     */
    private function getOrders(): array {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('Y-m-d\TH:i:s');

        $offset = $this->configService->get('June.config.orderOffset') ?? 0;

        $orderCriteria = new Criteria();
        /** @DEV
        $orderCriteria->addFilter(new RangeFilter('orderDateTime',
            [
                RangeFilter::LTE => $installDate
            ]
        ));
         */
        /**@DEV: 24 Monate zurück gehen
        $now24MonthsAgo = new \DateTime();
        $now24MonthsAgo->setTimestamp($now24MonthsAgo->getTimestamp() - (60 * 60 * 24 * 365 * 2));
        $formatedUtcNoOffset = $now24MonthsAgo->format('Y-m-d\TH:i:s');
        $orderCriteria->addFilter(new RangeFilter('orderDateTime',
            [
                RangeFilter::GTE => $formatedUtcNoOffset
            ]
        ));
        */

        //@DEV: which orders should be ignored for mass-export?
        $orderCriteria->addAssociation('stateMachineState.technicalName');
        $orderCriteria->addFilter(new NotFilter(MultiFilter::CONNECTION_AND, [
            new EqualsAnyFilter('stateMachineState.technicalName', ['cancelled'])
        ]));
        //['completed', 'cancelled', 'in_progress', 'shipped', 'shipped_partially', 'returned', 'returned_partially', 'open']));


        $orderCriteria->setLimit(self::ORDER_LIMIT_PER_REQUEST);
        $orderCriteria->setOffset($offset);
        //@TODO: rework orders to directly use lineItem association
        //$orderCriteria->addAssociation('lineItems');
        //$orderCriteria->addAssociation('orderCustomer');
        $orderCriteria->addAssociation('orderCustomer.customer');
        $orderCriteria->addAssociation('currency');
        $orderCriteria->addAssociation('billingAddress.country');
        $orderCriteria->addAssociation('deliveries.shippingMethod.name');
        $orderCriteria->addAssociation('transactions.paymentMethod');

        return $this->orderRepository->search($orderCriteria, Context::createDefaultContext())->getElements();
    }

    private function getAvailableOrdersCount(): int {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('c');

        $orderCriteria = new Criteria();
        /** @DEV:
        $orderCriteria->addFilter(new RangeFilter('orderDateTime',
        [
            RangeFilter::LTE => $installDate
        ]
        ));
         */
        /**@DEV: 24 Monate zurück gehen
        $now24MonthsAgo = new \DateTime();
        $now24MonthsAgo->setTimestamp($now24MonthsAgo->getTimestamp() - (60 * 60 * 24 * 365 * 2));
        $formatedUtcNoOffset = $now24MonthsAgo->format('Y-m-d\TH:i:s');
        $orderCriteria->addFilter(new RangeFilter('orderDateTime',
            [
                RangeFilter::GTE => $formatedUtcNoOffset
            ]
        ));
        */

        //@DEV: which orders should be ignored for mass-export?
        $orderCriteria->addAssociation('stateMachineState.technicalName');
        $orderCriteria->addFilter(new NotFilter(MultiFilter::CONNECTION_AND, [
            new EqualsAnyFilter('stateMachineState.technicalName', ['cancelled'])
        ]));

        $offset = $this->configService->get('June.config.orderOffset') ?? 0;
        $orderCriteria->setLimit(self::ORDER_LIMIT_PER_REQUEST);
        $orderCriteria->setOffset($offset);

        return $this->orderRepository->search($orderCriteria, Context::createDefaultContext())->count();
    }

    private function getTotalOrdersCount(): int {
        $pluginCriteria = new Criteria();
        $pluginCriteria = $pluginCriteria->addFilter(new EqualsFilter('name', 'June'));
        $plugin = $this->pluginRepository->search($pluginCriteria, Context::createDefaultContext())->first();
        $installDate = $plugin->getInstalledAt()->format('c');

        $orderCriteria = new Criteria();
        /**@DEV:
        $orderCriteria->addFilter(new RangeFilter('orderDateTime',
            [
                RangeFilter::LTE => $installDate
            ]
        ));
         */
        /**@DEV: 24 Monate zurück gehen
        $now24MonthsAgo = new \DateTime();
        $now24MonthsAgo->setTimestamp($now24MonthsAgo->getTimestamp() - (60 * 60 * 24 * 365 * 2));
        $formatedUtcNoOffset = $now24MonthsAgo->format('Y-m-d\TH:i:s');
        $orderCriteria->addFilter(new RangeFilter('orderDateTime',
            [
                RangeFilter::GTE => $formatedUtcNoOffset
            ]
        ));
        */

        //@DEV: which orders should be ignored for mass-export?
        $orderCriteria->addAssociation('stateMachineState.technicalName');
        $orderCriteria->addFilter(new NotFilter(MultiFilter::CONNECTION_AND, [
            new EqualsAnyFilter('stateMachineState.technicalName', ['cancelled'])
        ]));

        return $this->orderRepository->search($orderCriteria, Context::createDefaultContext())->count();
    }

    /**
     * @param OrderEntity $order
     * @return array<OrderLineItemEntity>
     */
    private function getLineItems(OrderEntity $order): array {
        $itemCriteria = new Criteria();
        $itemCriteria->addFilter(new EqualsFilter("orderId", $order->getId()));
        $itemCriteria->addAssociation('cover');
        return $this->orderLineItemsRepository->search($itemCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * @param $productId
     * @return ProductEntity|null
     */
    private function getProduct($productId): ?ProductEntity {
        if ($productId) {
            $itemCriteria = new Criteria([$productId]);
            $itemCriteria->addAssociation('cover');
            return $this->productRepository->search(new $itemCriteria, Context::createDefaultContext())->first();
        }

        return null;
    }

    /**
     * @param $mediaId
     * @return ProductMediaEntity|null
     */
    private function getProductCover($mediaId): ?ProductMediaEntity {
        return ($mediaId ? $this->productMediaRepository->search(new Criteria([$mediaId]), Context::createDefaultContext())->first() : null);
    }

    /**
     * @param OrderEntity $order
     * @return LocaleEntity|null
     */
    private function getLocale(OrderEntity $order): ?LocaleEntity {
        $locale = null;

        if ($order->getLanguageId()) {
            $languageCriteria = new Criteria();
            $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $order->getLanguageId()));
            $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

            if ($language->getLocaleId()) {
                $localeCriteria = new Criteria();
                $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
                $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();
            }
        }

        return $locale;
    }

    /**
     * @param OrderEntity $order
     * @return CountryEntity|null
     */
    private function getCountry(OrderEntity $order): ?CountryEntity {
        $country = null;

        if ($order->getSalesChannelId()) {
            $salesChannelCriteria = new Criteria();
            $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $order->getSalesChannelId()));
            $salesChannel = $this->salesRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

            if ($salesChannel->getCountryId()) {
                $countryCriteria = new Criteria();
                $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
                $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();
            }
        }

        return $country;
    }

    /**
     * @param OrderCustomerEntity $orderCustomer
     * @return CustomerEntity|null
     */
    private function getCustomer(OrderCustomerEntity $orderCustomer, string $salesChannelId): ?CustomerEntity {
        if ($orderCustomer) {
            if ($orderCustomer->getCustomerId()) {
                return $this->customerEntity->search(new Criteria([$orderCustomer->getCustomerId()]), Context::createDefaultContext())->first();
            } else if ($orderCustomer->getCustomerNumber()) {
                $itemCriteria = new Criteria();
                $itemCriteria->addFilter(new EqualsFilter("customerNumber", $orderCustomer->getCustomerNumber()));
                $itemCriteria->addFilter(new EqualsFilter("salesChannelId", $salesChannelId));
                $itemCriteria->addAssociation('tags');
                // $itemCriteria->addAssociation('group');
                return $this->customerEntity->search($itemCriteria, Context::createDefaultContext())->first();
            }
        }
        return null;
    }

    /**
     * @param string $tagId
     * @return string
     */
    private function getTag(string $tagId) :string {
        if (!in_array($tagId, $this->loadedCustomerTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedCustomerTags[$tagId] = $tag->getName();
        }

        return $this->loadedCustomerTags[$tagId];
    }

    /**
     * @param OrderCustomerEntity|null $orderCustomer
     * @return NewsletterRecipientEntity|null
     */
    private function getNewsletterSubscriber(?OrderCustomerEntity $orderCustomer): ?NewsletterRecipientEntity {
        if ($orderCustomer && $orderCustomer->getEmail()) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $orderCustomer->getEmail()));
            $itemCriteria->addAssociation('tags');
            return $this->newsletterRecipientEntity->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }
}
