<?php declare(strict_types=1);

namespace June\Service\ScheduledTask;

use June\Components\Api\Client\JuneApiClient;
use June\Components\Logger\Helper\LoggerHelper;
use Psr\Log\LoggerInterface;
use Shopware\Core\Checkout\Customer\CustomerEntity;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\OrFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Sorting\FieldSorting;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTask;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskDefinition;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskEntity;
use Shopware\Core\Framework\MessageQueue\ScheduledTask\ScheduledTaskHandler;
use Shopware\Core\System\SalesChannel\SalesChannelEntity;
use Shopware\Core\System\SystemConfig\SystemConfigService;
use Shopware\Core\Content\Newsletter\Aggregate\NewsletterRecipient\NewsletterRecipientEntity;
use June\Subscriber\JuneCommerceSubscriber;

use Shopware\Core\System\Locale\LocaleEntity;
use Shopware\Core\System\Country\CountryEntity;
use Shopware\Core\System\Tag\TagEntity;
use Symfony\Component\Messenger\Attribute\AsMessageHandler;

#[AsMessageHandler(handles: ExportSubscriberUpdatesTask::class)]
class ExportSubscriberUpdatesTaskHandler extends ScheduledTaskHandler {

    const SUBSCRIBER_LIMIT_PER_REQUEST = 200;
    const CONF_CURSOR_DATE_POSITION = 'June.config.last_subscriber_updated_at_cursor';
    const CONF_CURSOR_PAGINATION_POSITION = 'June.config.last_subscriber_pagination_cursor';

    private EntityRepository $languageRepository;

    private EntityRepository $localeRepository;

    private EntityRepository $countryRepository;

    private EntityRepository $newsletterSubscriberRepository;

    private EntityRepository $customerRepository;

    private EntityRepository $customerTagsRepository;

    private EntityRepository $salesChannelRepository;

    private SystemConfigService $configService;

    private LoggerHelper $loggerHelper;

    private JuneApiClient $juneApiClient;

    private ?string $lastExecutionTime = null;

    private int $maxExecutionTime = 45;

    private array $loadedSubscriberTags = [];

    public function __construct
    (
        EntityRepository $scheduledTaskRepository,
        EntityRepository $languageRepository,
        EntityRepository $localeRepository,
        EntityRepository $countryRepository,
        EntityRepository $newsletterSubscriberRepository,
        EntityRepository $customerRepository,
        EntityRepository $customerTagsRepository,
        EntityRepository $salesChannelRepository,
        SystemConfigService $configService,
        LoggerHelper $loggerHelper,
        JuneApiClient $juneApiClient,
        LoggerInterface $logger,
    )
    {
        parent::__construct($scheduledTaskRepository, $logger);
        $this->languageRepository = $languageRepository;
        $this->localeRepository = $localeRepository;
        $this->countryRepository = $countryRepository;
        $this->newsletterSubscriberRepository = $newsletterSubscriberRepository;
        $this->customerRepository = $customerRepository;
        $this->customerTagsRepository = $customerTagsRepository;
        $this->salesChannelRepository = $salesChannelRepository;
        $this->configService = $configService;
        $this->loggerHelper = $loggerHelper;
        $this->juneApiClient = $juneApiClient;

        $serverMaxExecutionTime = ini_get('max_execution_time');
        if ($serverMaxExecutionTime && $serverMaxExecutionTime > 0) {
            $this->maxExecutionTime = $serverMaxExecutionTime - 5;
            if ($this->maxExecutionTime <= 0) {
                $this->maxExecutionTime = $serverMaxExecutionTime * 1;
            }
        }

        $this->lastExecutionTime = null;
    }

    /**
     * @param ScheduledTask $task
     * @return void
     */
    public function __invoke(ScheduledTask $task): void
    {
        $taskId = $task->getTaskId();
        if ($taskId === null) {
            // run task independent of the schedule
            $this->run();
            return;
        }

        /** @var ScheduledTaskEntity|null $taskEntity */
        $taskEntity = $this->scheduledTaskRepository
            ->search(new Criteria([$taskId]), Context::createDefaultContext())
            ->get($taskId);

        if ($taskEntity === null || !$taskEntity->isExecutionAllowed()) {
            return;
        }

        $this->markTaskRunning($task);

        try {
            $lastTaskExecution = $taskEntity->getLastExecutionTime();

            if ($this->configService->get(self::CONF_CURSOR_DATE_POSITION)) {
                $this->lastExecutionTime = $this->configService->get(self::CONF_CURSOR_DATE_POSITION);
            } else if ($lastTaskExecution) {
                $this->lastExecutionTime = $lastTaskExecution->format('c');
            } else {
                $now = new \DateTime('now');
                $minutes10Ago = new \DateTime('now');
                $minutes10Ago->setTimestamp($now->getTimestamp()-(60*10));
                $this->lastExecutionTime = $minutes10Ago->format('c');
            }
            //$this->loggerHelper->logInfo("starting exportSubscriberUpdates. Last execution: ".$this->lastExecutionTime, "exportSubscriberUpdates");

            $this->run();
        } catch (\Throwable $e) {
            $this->markTaskFailed($task);
            $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportSubscriberUpdates");
            return;
        }

        $this->rescheduleTask($task, $taskEntity);
    }

    /**
     * @return iterable
     */
    public static function getHandledMessages(): iterable
    {
        return [ExportSubscriberUpdatesTask::class];
    }

    /**
     * @return void
     */
    public function run(): void
    {

        try {
            $customerData = [];
            $customerUpdateFields = [
                "external_id",
                "email",
                "firstname",
                "lastname",
                "salutation",
                "salutation_readable",
                "store_id",
                "locale",
                "created_at",
                "updated_at",
                "geo_location",
                "guest"
            ];

            $allSalesChannels = $this->getAllSalesChannels();

            $subscribers = $this->getNewsletterSubscribers();
            if ($subscribers && count($subscribers)) {
                $limitReached = count($subscribers)>=self::SUBSCRIBER_LIMIT_PER_REQUEST;
                $lastUpdatedDate = null;
                foreach ($subscribers as $subscriber) {
                    if (!$subscriber || !$subscriber->getId() || !$subscriber->getEmail()) {
                        continue;
                    }
                    $lastUpdatedDate = $subscriber->getUpdatedAt();

                    $salesChannelId = $subscriber->getSalesChannelId();
                    $customer = $this->getCustomer($subscriber, $salesChannelId);
                    $locale = $this->getLocale($subscriber);
                    $country = $this->getCountry($subscriber);

                    $enrichedCustomer = [
                        'external_id'               => $customer?->getCustomerNumber() ?? $subscriber->getId(),
                        'email'                     => strtolower($subscriber->getEmail()),
                        'firstname'                 => $customer?->getFirstname() ?? $subscriber->getFirstname(),
                        'lastname'                  => $customer?->getLastname() ?? $subscriber->getLastname(),
                        'salutation'                => $customer?->getSalutation()?->getSalutationKey() ?? $subscriber->getSalutation()?->getSalutationKey(),
                        'salutation_readable'       => $customer?->getSalutation()?->getLetterName() ?? $subscriber->getSalutation()?->getLetterName(),
                        'store_id'                  => $salesChannelId,
                        'locale'                    => $locale?->getCode(),
                        'created_at'                => $customer?->getCreatedAt()?->format('c') ?? $subscriber->getCreatedAt()?->format('c'),
                        'updated_at'                => $customer?->getUpdatedAt()?->format('c') ?? $subscriber->getUpdatedAt()?->format('c'),
                        'geo_location'              => [
                            'country'                   => $country?->getName(),
                        ],
                        'guest'                     => !$customer || $customer->getGuest()
                    ];

                    if ($customer?->getId()) {
                        $enrichedCustomer['customer_id'] = $customer->getId();
                        $customerUpdateFields[] = "customer_id";
                    }

                    // add more information to customer if available
                    if ($subscriber->getConfirmedAt()) {
                        $enrichedCustomer['subscription_confirmed_at'] = $subscriber->getConfirmedAt()->format('c');
                        $customerUpdateFields[] = 'subscription_confirmed_at';
                    }

                    if ($customer?->getCustomFields()) {
                        foreach ($customer->getCustomFields() as $fieldName => $fieldValue) {
                            if (!isset($enrichedCustomer[$fieldName])) {
                                $transformedFieldName = ($enrichedCustomer['guest'] ? 'guest_' : '') . $fieldName;
                                $enrichedCustomer[$transformedFieldName] = $fieldValue;

                                if (!in_array($transformedFieldName, $customerUpdateFields)) {
                                    $customerUpdateFields[] = $transformedFieldName;
                                }
                            }
                        }
                    }
                    if ($customer?->getBirthday()) {
                        $enrichedCustomer['birthday'] = $customer->getBirthday()->format('c');
                        if (!in_array('birthday', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'birthday';
                        }
                    }
                    if ($customer?->getLastLogin()) {
                        $enrichedCustomer['last_login'] = $customer->getLastLogin()->format('c');
                        if (!in_array('last_login', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_login';
                        }
                    }
                    if ($customer?->getLastOrderDate()) {
                        $enrichedCustomer['last_order_date'] = $customer->getLastOrderDate()->format('c');
                        if (!in_array('last_order_date', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'last_order_date';
                        }
                    }
                    if ($customer?->getOrderCount()) {
                        $enrichedCustomer['order_count'] = $customer->getOrderCount();
                        if (!in_array('order_count', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_count';
                        }
                    }
                    if ($customer?->getOrderTotalAmount()) {
                        $enrichedCustomer['order_total_amount'] = $customer->getOrderTotalAmount();
                        if (!in_array('order_total_amount', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'order_total_amount';
                        }
                    }

                    if ($customer?->getCampaignCode()) {
                        $enrichedCustomer['campaign_code'] = $customer->getCampaignCode();
                        if (!in_array('campaign_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'campaign_code';
                        }
                    }

                    if ($customer?->getAffiliateCode()) {
                        $enrichedCustomer['affiliate_code'] = $customer->getAffiliateCode();
                        if (!in_array('affiliate_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'affiliate_code';
                        }
                    }

                    if ($customer?->getGroup()?->getName()) {
                        $enrichedCustomer['customer_group'] = $customer->getGroup()->getName();
                        if (!in_array('customer_group', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'customer_group';
                        }
                    }

                    $enrichedCustomer['_unsubscribed'] = !($subscriber->getStatus() === "direct" || $subscriber?->getStatus() === "optIn");
                    if (!in_array('_unsubscribed', $customerUpdateFields)) {
                        $customerUpdateFields[] = '_unsubscribed';
                    }

                    $enrichedCustomer['intent'] = $subscriber->getStatus();
                    if (!in_array('intent', $customerUpdateFields)) {
                        $customerUpdateFields[] = 'intent';
                    }

                    if ($subscriber->getCity()) {
                        $enrichedCustomer['city'] = $subscriber->getCity();
                        $enrichedCustomer['geo_location']['city'] = $subscriber->getCity();
                        if (!in_array('city', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'city';
                        }
                    }
                    if ($subscriber->getZipCode()) {
                        $enrichedCustomer['zip_code'] = $subscriber->getZipCode();
                        if (!in_array('zip_code', $customerUpdateFields)) {
                            $customerUpdateFields[] = 'zip_code';
                        }
                    }

                    if ($customer?->getTagIds()) {
                        $enrichedCustomer['_tags'] = [];
                        foreach($customer->getTagIds() AS $tagId) {
                            if ($tag = $this->getTag($tagId)) {
                                $enrichedCustomer['_tags'][] = $tag;
                            }
                        }
                        if (!in_array('_tags', $customerUpdateFields)) {
                            $customerUpdateFields[] = '_tags';
                        }
                    }
                    if ($subscriber->getTags()) {
                        if (!isset($enrichedCustomer['_tags'])) {
                            $enrichedCustomer['_tags'] = [];
                        }
                        foreach ($subscriber->getTags() as $tag) {
                            if ($tag?->getName() && !in_array($tag->getName(), $enrichedCustomer['_tags'])) {
                                $enrichedCustomer['_tags'][] = $tag->getName();
                            }
                        }
                        if (!in_array('_tags', $customerUpdateFields)) {
                            $customerUpdateFields[] = '_tags';
                        }
                    }

                    // if customer is assigned to single salesChannel, we define it for Export
                    if ($salesChannelId) {
                        if (!isset($customerData[$salesChannelId])) {
                            $customerData[$salesChannelId] = [];
                        }
                        $customerData[$salesChannelId][$subscriber->getEmail()] = $enrichedCustomer;
                    }
                    // if no salesChannels are assigned we loop through all channels
                    else if (!empty($allSalesChannels)) {
                        foreach ($allSalesChannels AS $singleSalesChannel) {
                            if (!isset($customerData[$singleSalesChannel->getId()])) {
                                $customerData[$singleSalesChannel->getId()] = [];
                            }

                            $enrichedCustomer['store_id'] = $singleSalesChannel->getId();
                            $customerData[$singleSalesChannel->getId()][$subscriber->getEmail()] = $enrichedCustomer;
                        }
                    }
                }

                // send collected data to June
                if (!empty($customerData)) {
                    foreach ($customerData AS $salesChannelId => $customers) {
                        $commerceCollectToken = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_TOKEN, $salesChannelId);
                        $commerceCollectAuth = $this->configService->get(JuneCommerceSubscriber::JUNE_COMMERCE_COLLECT_AUTH, $salesChannelId);

                        if($this->configService->get(JuneCommerceSubscriber::JUNE_TRANSFER_CUSTOMERS, $salesChannelId) && !empty($commerceCollectToken) && !empty($commerceCollectAuth)) {
                            // recreate comparison date for loading users to log it
                            $formatedUtcNoOffset = new \DateTime($this->lastExecutionTime);
                            $formatedUtcNoOffset = $formatedUtcNoOffset->format('Y-m-d\TH:i:s');

                            // log debug message
                            $this->loggerHelper->logInfo("Scheduled subscribers update export started for salesChannel '".$salesChannelId."' containing " . count($customers) . " subscribers created/updated after $formatedUtcNoOffset.", "exportSubscriberUpdates");

                            $requestData = [
                                "action" => "subscriber_update",
                                "json" => array_values($customers),
                                "activate_trigger" => true,
                                "default_subscribe" => false,
                                "force_subscribe" => false,
                                "update_if_exists" => true,
                                "overwrite_tags" => true,
                                "update_fields" => $customerUpdateFields
                            ];

                            $this->juneApiClient->sendData($requestData, "https://engagement.juneapp.com/collect/$commerceCollectToken/import/json", $commerceCollectAuth, $this->maxExecutionTime);
                        } else {
                            if (empty($commerceCollectToken)) {
                                $this->loggerHelper->logInfo("Scheduled subscribers update export for salesChannel '".$salesChannelId."' could not run because CommerceCollectToken is not configured in the JUNE extension configuration.", "exportSubscriberUpdates");
                            } elseif (empty($commerceCollectAuth)) {
                                $this->loggerHelper->logInfo("Scheduled subscribers update export for salesChannel '".$salesChannelId."' could not run because CommerceCollectAUTH is not configured in the JUNE extension configuration.", "exportSubscriberUpdates");
                            }
                        }
                    }
                }

                if ($limitReached) {
                    // increment pagination for next run
                    $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) + 1);
                    $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $this->lastExecutionTime);
                } else {
                    // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                    $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                    $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
                }
            }
            // no subscribers found in this execution
            else {
                // remove updatedAtDate for next execution (then uses last scheduleExecutionDate)
                $this->configService->set(self::CONF_CURSOR_PAGINATION_POSITION, 0);
                $this->configService->set(self::CONF_CURSOR_DATE_POSITION, null);
            }
        } catch (\Exception $e) {
            // set last execution for retry
            $this->configService->set(self::CONF_CURSOR_DATE_POSITION, $this->lastExecutionTime);

            $this->loggerHelper->logError($e->getMessage()." - ".$e->getTraceAsString(), "exportSubscriberUpdates");
        }
    }

    /**
     * @return array<SalesChannelEntity>
     */
    private function getAllSalesChannels(): array {
        $channelActiveFilter = new Criteria();
        $channelActiveFilter->addFilter(new EqualsFilter("active", true));
        return $this->salesChannelRepository->search($channelActiveFilter, Context::createDefaultContext())->getElements();
    }

    /**
     * @return array<NewsletterRecipientEntity>
     */
    private function getNewsletterSubscribers(): array {
        if (!$this->lastExecutionTime) {
            $now = new \DateTime('now');
            $minutes10Ago = new \DateTime('now');
            $minutes10Ago->setTimestamp($now->getTimestamp()-(60*10));
            $this->lastExecutionTime = $minutes10Ago->format('c');
        }

        $lastUpdatedSubscriberPage = $this->configService->get(self::CONF_CURSOR_PAGINATION_POSITION) ?? 0;

        // format updatedAt date and remove offset because Shopware adds it themselves
        $formatedUtcNoOffset = new \DateTime($this->lastExecutionTime);
        $formatedUtcNoOffset = $formatedUtcNoOffset->format('Y-m-d\TH:i:s');

        $subscriberCriteria = new Criteria();
        $subscriberCriteria->addFilter(new OrFilter(
            [
                new RangeFilter('createdAt',
                    [
                        RangeFilter::GTE => $formatedUtcNoOffset
                    ]
                ),
                new RangeFilter('updatedAt',
                    [
                        RangeFilter::GTE => $formatedUtcNoOffset
                    ]
                )
            ])
        );
        $subscriberCriteria->addSorting(new FieldSorting('updatedAt', FieldSorting::ASCENDING));
        if ($lastUpdatedSubscriberPage) {
            $subscriberCriteria->setOffset($lastUpdatedSubscriberPage * self::SUBSCRIBER_LIMIT_PER_REQUEST);
        }
        $subscriberCriteria->setLimit(self::SUBSCRIBER_LIMIT_PER_REQUEST);
        $subscriberCriteria->addAssociation('tags');
        $subscriberCriteria->addAssociation('salutation');

        // $this->loggerHelper->logInfo("Scheduled subscribers query: Offset:".$lastUpdatedSubscriberPage * self::SUBSCRIBER_LIMIT_PER_REQUEST ." - Limit:".self::SUBSCRIBER_LIMIT_PER_REQUEST ." - UpdatedAt>=".$formatedUtcNoOffset, "exportSubscriberUpdates");

        return $this->newsletterSubscriberRepository->search($subscriberCriteria, Context::createDefaultContext())->getElements();
    }

    /**
     * @param NewsletterRecipientEntity $subscriber
     * @param string $salesChannelId
     * @return CustomerEntity|null
     */
    private function getCustomer(NewsletterRecipientEntity $subscriber, string $salesChannelId): ?CustomerEntity {
        if ($subscriber->getEmail()) {
            $itemCriteria = new Criteria();
            $itemCriteria->addFilter(new EqualsFilter("email", $subscriber->getEmail()));
            $itemCriteria->addFilter(new EqualsFilter('salesChannelId', $salesChannelId));
            $itemCriteria->addAssociation('tags');
            $itemCriteria->addAssociation('group');
            return $this->customerRepository->search($itemCriteria, Context::createDefaultContext())->first();
        }
        return null;
    }

    /**
     * @param NewsletterRecipientEntity $subscriber
     * @return LocaleEntity|null
     */
    private function getLocale(NewsletterRecipientEntity $subscriber): ?LocaleEntity {
        $locale = null;

        if ($subscriber->getLanguageId()) {
            $languageCriteria = new Criteria();
            $languageCriteria = $languageCriteria->addFilter(new EqualsFilter('id', $subscriber->getLanguageId()));
            $language = $this->languageRepository->search($languageCriteria, Context::createDefaultContext())->first();

            if ($language->getLocaleId()) {
                $localeCriteria = new Criteria();
                $localeCriteria = $localeCriteria->addFilter(new EqualsFilter('id', $language->getLocaleId()));
                $locale = $this->localeRepository->search($localeCriteria, Context::createDefaultContext())->first();
            }
        }

        return $locale;
    }

    /**
     * @param NewsletterRecipientEntity $subscriber
     * @return CountryEntity|null
     */
    private function getCountry(NewsletterRecipientEntity $subscriber): ?CountryEntity {
        $country = null;

        if ($subscriber->getSalesChannelId()) {
            $salesChannelCriteria = new Criteria();
            $salesChannelCriteria = $salesChannelCriteria->addFilter(new EqualsFilter('id', $subscriber->getSalesChannelId()));
            $salesChannel = $this->salesChannelRepository->search($salesChannelCriteria, Context::createDefaultContext())->first();

            if ($salesChannel->getCountryId()) {
                $countryCriteria = new Criteria();
                $countryCriteria = $countryCriteria->addFilter(new EqualsFilter('id', $salesChannel->getCountryId()));
                $country = $this->countryRepository->search($countryCriteria, Context::createDefaultContext())->first();
            }
        }

        return $country;
    }

    /**
     * @param string $tagId
     * @return string
     */
    private function getTag(string $tagId): string {
        if (!in_array($tagId, $this->loadedSubscriberTags)) {
            $tag = $this->customerTagsRepository->search(new Criteria([$tagId]), Context::createDefaultContext())->first();
            $this->loadedSubscriberTags[$tagId] = $tag->getName();
        }

        return $this->loadedSubscriberTags[$tagId];
    }
}
